#!/usr/bin/env python
'''
copyright 2010 jezra lickter http://www.jezra.net

heybuddy is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

heybuddy is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with heybuddy.  If not, see <http://www.gnu.org/licenses/>.
'''
import gtk,gobject
import sys, os
from xml.dom.minidom import parse, parseString
import re
import time
#what are the custom classes that I need?
from Communicator import Communicator
from Configuration import Configuration
from XMLProcessor import XMLProcessor
from ImageCache import ImageCache
from StatusIcon import StatusIcon
from About import About
from Settings import Settings
import MainWindow
global app_name
global version
global branch
app_name = "heybuddy"
version = "0.1.5"
branch = "Milton"
highest_id={MainWindow.DENT:0,MainWindow.MENTION:0,MainWindow.DIRECT:0}
#create a fake enum for options
class application :
	def __init__(self):
		#init threading
		gobject.threads_init()
		#keep track of the window size
		self.window_height=0
		self.window_width=0
		self.window_is_hidden = False
		self.has_status_icon = False
		#keep track of the filters
		self.filters ={'tags':[],'users':[],'groups':[]}
		self.options={'run_as_tray_app':False,'context_backwards':False,'no_avatars':False}
		#keep track of the direct_message requests
		self.last_direct_message_time=0
		self.waiting_requests=0
		self.is_first_dents = True
		#what are the assets?
		asset_dir = "assets"
		heybuddy_dir = os.path.dirname( os.path.realpath( __file__ ) )
		self.readme_file  = os.path.join(heybuddy_dir,"README.txt")
		self.default_icon_path = os.path.join(heybuddy_dir,asset_dir,"icon.png")
		self.networking_icon_path = os.path.join(heybuddy_dir,asset_dir,"icon1.png")
		self.throbber_icon_path = os.path.join(heybuddy_dir,asset_dir,"throbber.gif")
		self.conf = Configuration(app_name)
		self.xmlprocessor = XMLProcessor()
		#create a regex to replace parts of a dent
		self.regex_tag = re.compile("(^| )(#[A-z0-9_]+)")
		self.regex_group = re.compile("(^| )(![A-z0-9_]+)")
		self.regex_user=re.compile("(^| )(@[A-z0-9_]+)")
		self.regex_url=re.compile("(http[s]?://.*?)(\s|$)",re.IGNORECASE)
		#create filter regex
		self.regex_filter_tags = re.compile("#([A-z0-9_]+)")
		self.regex_filter_groups = re.compile("!([A-z0-9_]+)")
		#create the communicator
		self.comm = Communicator(app_name)
		self.comm.connect('statusesXML',self.process_statusesXML,MainWindow.DENT)
		self.comm.connect('mentionsXML',self.process_statusesXML,MainWindow.MENTION)
		self.comm.connect('group-statusesXML',self.process_statusesXML,MainWindow.GROUP)
		self.comm.connect('user-statusesXML',self.process_statusesXML,MainWindow.USER)
		self.comm.connect('direct_messagesXML',self.process_statusesXML,MainWindow.DIRECT,True)
		self.comm.connect('conversationXML',self.process_conversationXML)
		self.comm.connect('userXML',self.process_userXML)
		self.comm.connect('groupXML',self.process_groupXML)
		self.comm.connect('new-statusXML',self.process_new_statusXML)
		self.comm.connect('user_is_friendXML',self.process_user_is_friendXML)
		self.comm.connect('user_is_memberXML',self.process_user_is_memberXML)
		self.comm.connect('friendshipXML',self.process_friendshipXML)
		self.comm.connect('exception-caught',self.communication_exception)
		self.comm.connect('widget-image',self.process_widget_image)
		#create an image cacher thingy
		self.imagecache = ImageCache()
		self.imagecache.set_disabled( self.conf.get_bool_option("no_avatars") )
		self.imagecache.connect('get-widget-image', self.get_widget_image)
		#build the gui
		self.build_gui()
		#align the window
		self.align_window()
		# do we need to create the status icon
		if self.conf.get_bool_option('run_as_tray_app'):
			self.create_status_icon()
		
	def build_gui(self):
		#build the GUI 
		self.mainwindow=MainWindow.MainWindow(app_name,version,branch)
		try:
			self.mainwindow.set_icon_from_file(self.default_icon_path)
		except:
			print "failed to load %s" % (icon_path)
		self.mainwindow.connect("destroy", self.quit_app )
		self.mainwindow.connect("update-status",self.update_status)
		self.mainwindow.connect("get-conversation",self.get_conversation)
		self.mainwindow.connect("show-user",self.get_user_info)
		self.mainwindow.connect("show-group",self.get_group_info)
		self.mainwindow.connect("notebook-page-change",self.page_change)
		self.mainwindow.connect('follow-user',self.follow_user)
		self.mainwindow.connect('join-group',self.join_group)
		self.mainwindow.connect('add-tag-filter',self.add_tag_filter)
		self.mainwindow.connect('remove-tag-filter',self.remove_tag_filter)
		self.mainwindow.connect('add-group-filter', self.add_group_filter )
		self.mainwindow.connect('remove-group-filter', self.remove_group_filter )
		self.mainwindow.connect('add-user-filter', self.add_user_filter )
		self.mainwindow.connect('remove-user-filter', self.remove_user_filter )
		#keep track of the window size
		self.mainwindow.connect('size-allocate', self.window_size_allocate)
		self.mainwindow.connect('window-state-event', self.mainwindow_state_event)
		self.mainwindow.connect('delete-event', self.mainwindow_delete_event)
		#create and add the account page
		self.settings = Settings()
		#update the configuration of the settings
		self.settings.connect("update-account",self.update_account)
		self.options['run_as_tray_app'] = self.conf.get_bool_option('run_as_tray_app')
		self.settings.set_run_as_tray_app( self.options['run_as_tray_app']  )
		self.options['context_backwards']  = self.conf.get_bool_option('context_backwards')
		self.settings.set_context_backwards( self.options['context_backwards'] )
		self.options['no_avatar']= self.conf.get_bool_option('no_avatars')
		self.settings.set_no_avatars( self.options['no_avatar'] )
		#connect the setting signals
		self.settings.connect('option-run-as-tray-app', self.option_changed, 'run_as_tray_app')
		self.settings.connect('option-context-backwards',self.option_changed, 'context_backwards')
		self.settings.connect('option-no-avatars',self.option_changed, 'no_avatars')
		#add the settings to the mainwindow
		self.mainwindow.add_notebook_page(self.settings,gtk.Label("Settings"))
		#create and add the about page
		about = About(version,branch,self.default_icon_path,self.readme_file)
		self.mainwindow.add_notebook_page( about, gtk.Label("About") )
		self.mainwindow.show_some()
		
	def align_window(self):
		#try to set the mainwindows size based on conf info
		try:
			w = int( self.conf.get('window_info','width') )
			h = int( self.conf.get('window_info','height') )
			self.mainwindow.resize(w,h)
		except:
			pass
		try:
			x = int( self.conf.get('window_info','x') )
			y = int( self.conf.get('window_info','y') )
			self.mainwindow.move(x,y)
		except:
			pass
		
		#get the filters
		try: 
			#read the filters
			filters = self.conf.get('settings','filters',True)
			#the filters should be an array
			if len(filters):
				self.filters = filters
				#set the filters
				self.mainwindow.set_tag_filters( self.filters['tags'] )
				self.mainwindow.set_user_filters( self.filters['users'] )
				self.mainwindow.set_group_filters( self.filters['groups'] )
		except:
			pass
	
	def create_status_icon(self):
		if not self.has_status_icon:
			#create the statusicon
			self.statusicon = StatusIcon()
			self.statusicon.connect('activate', self.status_clicked )
			self.statusicon.set_icon( self.default_icon_path )
		self.statusicon.set_property("visible",True)
		self.has_status_icon=True
	
	def destroy_status_icon(self):
		self.statusicon.set_property("visible",False)
		self.has_status_icon=False
		
	def update_account(self,widget,n,p):
		if n!=self.conf.name() or p!=self.conf.password():
			#TODO:check if these are valid
			#self.comm.check_validity(n,p)
			self.valid_account_info(n,p)
			self.mainwindow.message_label.hide()
			
	def increment_requests(self):
		self.waiting_requests+=1
		if self.waiting_requests==1:
			try:
				self.mainwindow.set_icon_from_file(self.networking_icon_path)
			except:
				print "failed to load %s" % (self.networking_icon_path)
			try:
				self.mainwindow.set_throbber(self.throbber_icon_path)
			except:
				print "failed to load %s" % (self.throbber_icon_path)
			if self.has_status_icon:
				self.statusicon.set_icon( self.networking_icon_path)
				
	def decrement_requests(self):
		self.waiting_requests-=1
		if self.waiting_requests==0:
			try:
				self.mainwindow.set_icon_from_file(self.default_icon_path)
			except:
				print "failed to load %s" % (self.default_icon_path)
			try:
				self.mainwindow.set_throbber(self.default_icon_path)
			except:
				print "failed to load %s" % (self.default_icon_path)
			if self.has_status_icon:
				self.statusicon.set_icon( self.default_icon_path )
		if self.waiting_requests<0:
			self.waiting_requests=0
	
	def valid_account_info(self,n,p):
		self.conf.name( n )
		self.conf.password( p )
		self.conf.save()
		#update the comm
		self.comm.set_name_and_password( n,p )
		#switch to the dents page
		self.mainwindow.set_notebook_page(MainWindow.DENT) 
		self.get_statuses()

	def get_statuses(self):
		self.increment_requests()
		self.comm.get_statuses(highest_id[MainWindow.DENT])
	
	def get_mentions(self):
		self.increment_requests()
		self.comm.get_mentions(highest_id[MainWindow.MENTION])	
	
	def get_direct_messages(self):
		now = time.time()
		#has it been 3 minutes since the last request?
		if now-(3*60) > self.last_direct_message_time:
			self.increment_requests()
			self.comm.get_direct_messages(highest_id[MainWindow.DIRECT])	
			self.last_direct_message_time=now
	
	def run(self):
		#start the communicator
		self.comm.start()
		self.main_loop = gobject.MainLoop()
		#add a function to the main loop
		gobject.timeout_add(90*1000, self.update_statuses )
		gobject.timeout_add(3*60*1000, self.update_mentions)
		#we have to temporarily show the various pages or nothing works
		self.mainwindow.set_notebook_page(MainWindow.MENTION)
		self.mainwindow.set_notebook_page(MainWindow.DENT)
		#did the conf read a name and password?
		if( self.conf.name() !="" and self.conf.password()!=""):
			self.comm.set_name_and_password( self.conf.name(),self.conf.password() )
			#fill the Account tab name/pw fields to show the user that they're logged in
			self.settings.set_name(self.conf.name())
			self.settings.set_password(self.conf.password())
			#add a timout to get dents
			gobject.timeout_add(500,self.get_statuses )
			gobject.timeout_add(600,self.get_mentions )
			#gobject.timeout_add(1000,self.get_direct_messages )
		else:
			#set the focus on the account page; the last page
			self.mainwindow.set_notebook_page(MainWindow.ACCOUNT)
		
		self.mainwindow.run()
		self.main_loop.run()

	def update_statuses(self):
		self.get_statuses()
		return True
	
	def update_mentions(self):
		self.get_mentions()
		return True
	
	def update_direct_messages(self):
		self.get_direct_messages()
		return True
	
	def update_status(self,widget,text,id):
		self.increment_requests()
		#remove newlines from text
		text=text.replace("\n","")
		self.comm.send_update_status(text,id)
		
	def quit_app(self,widget=None):
		self.save_window_info()
		#quit the communicator loop
		self.comm.quit()
		#quit the main loop
		self.main_loop.quit()
		
	
	def get_widget_image(self,imagecache,image):
		self.comm.get_widget_image(image)
	
	def get_conversation(self,window=None,id=0):
		self.increment_requests()
		self.comm.get_conversation(id)
	
	def get_user_info(self,widget,name):
		self.increment_requests()
		self.comm.get_user_info(name)
		self.increment_requests()
		self.comm.get_user_is_friend(name)
		self.increment_requests()
		self.comm.get_user_statuses(name)
	
	def get_group_info(self,widget,name):
		self.increment_requests()
		self.comm.get_group_info(name)
	
	def follow_user(self,widget,name,bool):
		if bool:
			self.comm.friendship_create(name)
		else:
			self.comm.friendship_destroy(name)
	
	def join_group(self,widget,name,bool):
		if bool:
			self.comm.group_join(name)
		else:
			self.comm.group_leave(name)
	
	def process_friendshipXML(self,widget,text):
		pass
	
	def page_change(self,widget,old_id,new_id):
		if new_id==MainWindow.DIRECT:
			self.get_direct_messages()
	
	def process_userXML(self,comm,text):
		self.decrement_requests()
		data = self.xmlprocessor.get_user_info(text)
		self.mainwindow.display_user(data)
	
	def process_groupXML(self,comm,text):
		self.decrement_requests()
		data = self.xmlprocessor.get_group_info(text)
		if data['stream_logo']!=None:
			self.imagecache.add_image_to_widget(data['stream_logo'],self.mainwindow.grouppage)
		#is this user a member of this group
		self.increment_requests()
		self.comm.get_user_is_member(data['id'])
		self.increment_requests()
		self.comm.get_group_statuses(data['id'])
		self.mainwindow.display_group(data)
	
	def process_user_is_friendXML(self,comm,text):
		self.decrement_requests()
		is_friend = self.xmlprocessor.get_user_is_friend(text)
		self.mainwindow.display_user_is_friend( is_friend )
	
	def process_user_is_memberXML(self,comm,text):
		self.decrement_requests()
		is_member = self.xmlprocessor.get_user_is_member(text)
		self.mainwindow.display_user_is_member( is_member )
	
	def process_new_statusXML(self,comm,text):
		#a update status was sent, get all new statuses
		#self.update_statuses()	
		self.decrement_requests()
		#there is one statusa and it is the DOM
		status = self.xmlprocessor.get_dom(text)
		data = self.xmlprocessor.get_dent_data(status)
		data['markup'] = self.markup_dent_text(data['text'])
		#add the dent
		dent = self.mainwindow.add_dent(data,MainWindow.DENT)
		self.imagecache.add_image_to_widget(data['profile_image_url'],dent)
	
	def process_conversationXML(self,object,text):
		self.decrement_requests()
		#get the status
		status = self.xmlprocessor.get_dom(text)
		data = self.xmlprocessor.get_dent_data(status)
		data['markup']=self.markup_dent_text(data['text'])
		#since this is a conversation, we can get rid of the in_reply_to_screen_name
		data['in_reply_to_screen_name']=None
		#tell the mainWindow to add the dent
		dent=self.mainwindow.add_dent(data,MainWindow.CONVERSATION,True, self.conf.get('options','context_backwards') )
		self.imagecache.add_image_to_widget(data['profile_image_url'],dent)
		if data['in_reply_to_id']!=None:
			#recursively get in_reply_to_ids
			self.get_conversation(id=data['in_reply_to_id'])
		
	def get_dent_time(self,text):
		print text
		
	def markup_dent_text(self,text):
		#process the text to markup
		#xmlencode the string
		text = self.xmlprocessor.xmlentities(text)
		#regex find the highlights
		markup = self.regex_tag.sub(r'\1<b>\2</b>',text)
		#regex find users
		if MainWindow.has_hildon:
			markup = self.regex_group.sub(r'\1<b>\2</b>',markup)
			markup = self.regex_user.sub(r'\1<b>\2</b>',markup)
		else:
			#find urls
			markup = self.regex_url.sub(r'<a href="\1">\1</a>\2',markup)
			markup = self.regex_user.sub(r'\1<a href="\2">\2</a>',markup)
			markup = self.regex_group.sub(r'\1<a href="\2">\2</a>',markup)
		return markup
	
	def process_statusesXML(self,object,text,target_index,is_direct=False):
		self.decrement_requests()
		#get a list of the statuses
		statuses = self.xmlprocessor.get_statuses(text,is_direct)
		#reverse the statuses list
		statuses.reverse()
		#what is the highest id of the target index?	
		for status in statuses:
			filtered_status = False
			data = self.xmlprocessor.get_dent_data(status)

			if target_index!=None:
				highest_id[target_index]=data['status_id']

			if target_index==MainWindow.DENT:
				#we should check for filtration
				filtered = self.check_filtered( data )
				filtered_status=filtered[0]
			if filtered_status:
				print "Filtered dent #%s matching %s" % (data['status_id'],filtered[1])
				#do nothing 
				pass
			else:
				#get the markup
				data['markup'] = self.markup_dent_text(data['text'])
				dent = self.mainwindow.add_dent(data,target_index)
				#get the image for this dent
				self.imagecache.add_image_to_widget(data['profile_image_url'],dent)
				#if the target_index = 0 and not first_dents and not is_conf
				if target_index==MainWindow.DENT and not self.is_first_dents:
					if "@"+self.conf.name() in data['markup']:					
						dent=self.mainwindow.add_dent(data,MainWindow.MENTION)
						self.imagecache.add_image_to_widget(data['profile_image_url'],dent)
					
		self.is_first_dents=False
		#get the related images
		self.imagecache.get_images()
	
	def process_widget_image(self,comm,data,name):
		self.imagecache.widget_image_add(data,name)	
		
	def communication_exception(self,communicator,code):
		self.decrement_requests()
		if code=="401": 
			#bad authorization
			self.mainwindow.set_message("Invalid Authorization: Your name or password is incorrect",MainWindow.ERROR_MESSAGE)
			self.mainwindow.set_notebook_page(MainWindow.ACCOUNT)
		elif code == "404":
			print 'not found'
		elif code == "503":
			print 'service unavailable'
		else:
			print 'unknown error: '+code
	
	#functions to handle the filtration system
	def add_tag_filter(self,winder,tag):
		if not tag in self.filters['tags']:
			self.filters['tags'].append(tag)
			self.filters['tags'].sort()
			self.mainwindow.set_tag_filters( self.filters['tags'] )
		
	def remove_tag_filter(self,winder,tag):
		if tag in self.filters['tags']:
			self.filters['tags'].remove(tag)
			self.mainwindow.set_tag_filters( self.filters['tags'] )
	
	def add_group_filter(self,winder,group):
		if not group in self.filters['groups']:
			self.filters['groups'].append(group)
			self.filters['groups'].sort()
			self.mainwindow.set_group_filters( self.filters['groups'] )
		
	def remove_group_filter(self,winder,group):
		if group in self.filters['groups']:
			self.filters['groups'].remove(group)
			self.mainwindow.set_group_filters( self.filters['groups'] )
			
	def add_user_filter(self,winder,user):
		if not user in self.filters['users']:
			self.filters['users'].append(user)
			self.filters['users'].sort()
			self.mainwindow.set_user_filters( self.filters['users'] )
		
	def remove_user_filter(self,winder,user):
		if user in self.filters['users']:
			self.filters['users'].remove(user)
			self.mainwindow.set_user_filters( self.filters['users'] )
	

	def check_filtered(self,data):
		if len(self.filters['tags'])==0 and len(self.filters['users'])==0 and len(self.filters['groups'])==0:
			return (False, None)
		#check the tags
		text = data['text']
		#what are the tags in the dent?
		dent_tags = self.regex_filter_tags.findall(text)
		for tag in dent_tags:
			for ftag in self.filters['tags']:
				if tag.lower()==ftag.lower():
					return (True, "tag: %s" % tag )
				
		#what are the groups in the dent?
		dent_groups = self.regex_filter_groups.findall(text)
		for group in dent_groups:
			for fgroup in self.filters['groups']:
				if group.lower()==fgroup.lower():
					return (True, "group: %s" % group )
				
		#check the user
		for user in self.filters['users']:
			if data['screen_name'].lower()==user.lower():
				return (True, "user: %s" % data['screen_name'] )
		
		return (False, None)
	
	def status_clicked(self,widget):
		if self.window_is_hidden==True:
			self.mainwindow.show_some()
			self.align_window()
			 			
	def window_size_allocate(self,widget,size):
		self.window_width = size[2]
		self.window_height = size[3]
			
	def save_window_info(self):
		#save the filters
		self.conf.set('settings','filters', self.filters ,True)
		#save the window info
		self.conf.set('window_info','width', self.window_width )
		self.conf.set('window_info','height', self.window_height )
		try:
			#some themes don't pass root origin?
			origin = self.mainwindow.window.get_root_origin()
			self.window_x = origin[0]
			self.window_y = origin[1]
			self.conf.set('window_info','x', self.window_x )
			self.conf.set('window_info','y', self.window_y )
		except:
			pass
		self.conf.save()
		
	def mainwindow_state_event(self,window, event):
		pass
		#print event.type
		
	def mainwindow_delete_event(self,window,event):
		self.save_window_info()
		#are we in tray mode?
		if self.conf.get_bool_option('run_as_tray_app') or self.has_status_icon:
			#is the status icon embedded?
			if self.statusicon.is_embedded():
				self.window_is_hidden=True
				self.mainwindow.hide_on_delete()
				return True
			
	def option_changed(self,widget,value,option):
		#save the option change in the configuration
		self.conf.set('options',option,value)
		self.options[option]=value
		self.conf.save()
		#is this the run_tray_app:
		if option=='run_as_tray_app':
			if value:
				self.create_status_icon()
			else:
				self.destroy_status_icon()
		elif option=='no_avatars':
			self.imagecache.set_disabled(value)

if(__name__=="__main__"):
	a = application()
	a.run()
